Object subclass: #ScratchProcess
	instanceVariableNames: 'stackFrame topBlock readyToYield errorFlag readyToTerminate blockVarList answer atomicityList tempVars scriptList isPaused timeOffset blockToDebug nextBlockToDebug doneWithDebugging '
	classVariableNames: 'BlockHighlightMSecs Times '
	poolDictionaries: ''
	category: 'Scratch-Execution Engine'!
!ScratchProcess commentStamp: '<historical>' prior: 0!
A ScratchProcess is what brings a stack of blocks to life. The process keeps track of which block to run next, evaluates block arguments, handles control structures, and so forth.

The ScratchFrameMorph is the scheduler, telling each process when to run by calling its runStep method. The runStep method will execute some number of blocks, then voluntarily yield control so that the ScratchFrameMorph can run another process. The etiquette is that a process should yield control at the end of every loop iteration, and while it is running a timed command (e.g. "wait 5 secs") or a synchronous command (e.g. "broadcast xxx and wait").

Structure:
  stackFrame		the ScratchStackFrame describing the current state of this process
  readyToYield		boolean indicating whether to yield control to another process
  errorFlag			boolean indicating whether an error was encountered
  readyToTerminate	boolean indicating whether the stop method has been called
!


!ScratchProcess methodsFor: 'initialization' stamp: 'jens 6/18/2010 01:55'!
initialize

	stackFrame _ nil.
	readyToYield _ false.
	errorFlag _ false.
	readyToTerminate _ false.
	isPaused _ false.
	doneWithDebugging _ false.
	tempVars _ VariableFrame new
! !


!ScratchProcess methodsFor: 'accessing' stamp: 'jens 6/11/2009 14:07'!
answer
	^answer! !

!ScratchProcess methodsFor: 'accessing' stamp: 'jm 11/20/2003 11:12'!
errorFlag
	"Answer true if this process has stopped due to an error."

	^ errorFlag
! !

!ScratchProcess methodsFor: 'accessing' stamp: 'jens 6/14/2009 16:28'!
errorFlag: aBoolean
	"Set the error flag for this process."

	| customBlocks |

	errorFlag _ aBoolean.
	stackFrame expression showError.
	stackFrame expression topBlock showErrorFeedback.

	"BYOB:"

	"find all custom blocks containing this error"

	customBlocks _ (stackFrame expression receiver customBlocksContaining: stackFrame expression).
	customBlocks isEmpty ifTrue: [^self ].

	topBlock allMorphs do: [:m|
		(customBlocks includes: m) ifTrue: [
			m showError ]].

	topBlock showErrorFeedback

! !

!ScratchProcess methodsFor: 'accessing' stamp: 'CWB 1/21/2013 18:09'!
expression: expression
	"Sets the expression to evaluate. This can be a single block or a collection of blocks. It should only be called before running the process."

	"Disabled in Enchanting -- we don't want to run code on the computer"
	^ self.
"
	stackFrame ifNotNil: [self error: 'Cannot modify expression'].
	stackFrame _ ScratchStackFrame new expression: expression.
"
! !

!ScratchProcess methodsFor: 'accessing' stamp: 'jens 6/15/2010 20:51'!
includesDeletedReceiver
	"Answer true if I contain a block whose receiver has been deleted."

	self stackAsArray do: [:b |

		(b isKindOf: CommandBlockMorph) ifTrue:[
			b receiver isNil
				ifTrue: [^ true ]
				ifFalse: [(b receiver isKindOf: VariableFrame)
					ifFalse: [b receiver owner isNil ifTrue: [
						^ true]]]]].

	^ false
! !

!ScratchProcess methodsFor: 'accessing' stamp: 'jm 5/28/2004 20:38'!
includesReceiver: anObject
	"Answer true if I have the given object as a block receiver."

	self stackAsArray do: [:frm |
		((frm respondsTo: #receiver) and: [frm receiver == anObject])
			ifTrue: [^ true]].
	^ false

! !

!ScratchProcess methodsFor: 'accessing' stamp: 'jm 6/4/2009 13:00'!
isRunning
	"Answer true if I am still running, have not encountered an error, and don't refer to any deleted objects."

	^ (stackFrame notNil and:
	   [errorFlag not and:
	   [self includesDeletedReceiver not]])
! !

!ScratchProcess methodsFor: 'accessing' stamp: 'jens 6/18/2010 02:02'!
popStackFrame
	"Pops the current stack frame off the top of the stack, so that the next one becomes the current frame."

	| command frame unlightFrame |

	frame _ stackFrame.
	stackFrame _ stackFrame parentFrame.

	command _ frame expression.

	"this effectively disables Scratch's so called single stepping mechanism
	for BYOB, a sacrifice that let's us catch up with Scratch speed again,
	-jens"

	true ifTrue: [^self].

	"and this is the old code"

	self isAtomic ifTrue: [^self ].

	BlockHighlightMSecs > 0 ifTrue: [
		(frame shouldUnlight not and: [command isKindOf: BlockMorph]) ifTrue: [
			"We push a new frame for unlighting this command."
			unlightFrame _ ScratchStackFrame new
				expression: command;
				shouldUnlight: true.
			self pushStackFrame: unlightFrame]].
! !

!ScratchProcess methodsFor: 'accessing' stamp: 'jm 12/30/2008 18:33'!
topBlock: aBlock

	topBlock _ aBlock.
! !


!ScratchProcess methodsFor: 'entry points' stamp: 'jens 7/8/2010 22:20'!
runStepFor: aStageMorph
	"Evaluates the current expression, then the next, etc., until the next yield."

	| expr isTimed |
	"evaluate blocks until we're ready to yield"

	isTimed _ stackFrame notNil and: [
		(stackFrame expression isKindOf: BlockMorph) and: [
			stackFrame expression isTimed]].

	(self isPaused and: [isTimed not]) ifTrue: [ ^ self].

	readyToYield _ false.
	[readyToYield or: [stackFrame isNil]] whileFalse: [
		self evaluateFor: aStageMorph].

	readyToTerminate ifTrue: [
		"pop all stack frames, unlighting enclosing blocks:"
		[stackFrame isNil] whileFalse: [
			expr _ stackFrame expression.
			(expr isKindOf: BlockMorph) ifTrue: [expr litUp: false].
			self popStackFrame]].

! !

!ScratchProcess methodsFor: 'entry points' stamp: 'jens 6/22/2010 21:55'!
stepUntilDoneFor: aStageMorph

	[self isRunning] whileTrue: [
		self isAtomic. "allow escaping out of forever loops in #showValue situations. -jens"
		Sensor keyboardPressed.
		self runStepFor: aStageMorph]
! !

!ScratchProcess methodsFor: 'entry points' stamp: 'jm 6/4/2009 13:12'!
stop
	"Permanently terminates this process."

	stackFrame ifNotNil: [stackFrame stopMIDI; stopMotors; stopTalkThinkAsk].
	readyToYield _ true.
	readyToTerminate _ true.
	topBlock ifNotNil: [topBlock scratchProc: nil].
! !


!ScratchProcess methodsFor: 'private-evaluation' stamp: 'jens 7/21/2010 01:02'!
applyPrimitive
	"Apply the current expression (which must be a CommandBlock) to the current arguments (which must all have been evaluated)."

	| value |

	value _ stackFrame expression evaluateWithArgs: stackFrame arguments.

	"save the return value in the parent frame before popStackFrame because popFrame adds a frame while single-stepping"
	self returnValueToParentFrame: value.
	self popStackFrame.

! !

!ScratchProcess methodsFor: 'private-evaluation' stamp: 'jens 7/8/2010 20:13'!
applyTimedCommand
	"Applies the current command to the already evaluated list of arguments over a particular time interval."

	| block arguments currentTime startTime args totalMSecs elapsedMSecs |
	block _ stackFrame expression.
	arguments _ stackFrame arguments.

	"Do we still need to evaluate more arguments?"
	arguments size < block argumentCount ifTrue: [^ self evaluateNextArgument].
	arguments _ block coerceArgs: arguments.

	"Record or get the time when command was first invoked."
	currentTime _ Time millisecondClockValue.
	startTime _ stackFrame startTime.
	startTime ifNil: [  "first call; just set starting time and value"
		timeOffset _ nil.
		args _ arguments asArray, (Array with: 0 with: nil).
		stackFrame startValue: (block receiver perform: block selector withArguments: args).
		stackFrame startTime: currentTime.
		readyToYield _ true.
		^ self].

	isPaused ifTrue: [
		timeOffset ifNil: [
			timeOffset _ currentTime - startTime].
		readyToYield _ true.
		^ self].

	timeOffset ifNotNil: [
		
		startTime _ currentTime - timeOffset.
		stackFrame startTime: startTime ]. "resume after pause -jens"

	timeOffset _ nil.

	"Call primitive time command with its arguments and the elapsed time in seconds"
	totalMSecs _ arguments last * 1000.
	block selector = #glideSecs:toX:y:elapsed:from: ifTrue: [totalMSecs _ arguments first * 1000].
	block selector = #mwait:elapsed:from: ifTrue: [totalMSecs _ arguments last].
	((block selector = #drum:duration:elapsed:from:) or:
	 [block selector = #noteOn:duration:elapsed:from:])
		ifTrue: [totalMSecs _ (60000 * arguments second) / block receiver tempo].
	block selector = #rest:elapsed:from:
		ifTrue: [totalMSecs _ (60000 * arguments first) / block receiver tempo].

	elapsedMSecs _ currentTime - startTime.

	currentTime < startTime ifTrue: [elapsedMSecs _ totalMSecs].  "clock wrap"
	args _ arguments asArray, (Array with: elapsedMSecs with: stackFrame startValue).
	block receiver perform: block selector withArguments: args.

	"If not done, then we leave stack as is and yield."
	elapsedMSecs < totalMSecs ifTrue: [
		readyToYield _ true.
		^ self].

	"Pop this command off the stack and return."
	self popStackFrame

! !

!ScratchProcess methodsFor: 'private-evaluation' stamp: 'jens 6/18/2010 01:58'!
evaluateCommandFor: aStageMorph
	"Evaluates the current block. If the argument is non-nil, redraw the stage."

	| expression |
	expression _ stackFrame expression.
	BlockHighlightMSecs > 1 ifTrue: [expression litUp: true].

	((blockToDebug notNil and: [blockToDebug ~= self currentBlock]) or: [
		self currentBlock = nextBlockToDebug]) ifTrue: [
			((ScratchFrameMorph someInstance viewMode = #presentation) | doneWithDebugging) ifFalse: [
				self pause.
				DebuggerFrameMorph new on: self]].

	expression isSpecialForm ifTrue: [^ self evaluateSpecialForm].

	stackFrame ifNil: [
		answer isNil ifFalse: [ ^self]].


	"evaluate arguments, if necessary"
	stackFrame arguments size < expression argumentCount
		ifTrue: [^ self evaluateNextArgument].

	expression isTimed ifTrue: [^ self applyTimedCommand].

	self applyPrimitive.

	aStageMorph ifNotNil: [
		aStageMorph updateTrailsForm.
		self isAtomic ifTrue: [^self].
		BlockHighlightMSecs = 1 ifTrue: [  "normal (non-turbo) mode; redraw after each cmd"
			World displayWorldSafely]].
! !

!ScratchProcess methodsFor: 'private-evaluation' stamp: 'jens 7/21/2010 19:49'!
evaluateFor: aStageMorph
	"Evaluates the current expression in the current environment for the given StageMorph."

	| expression |
	stackFrame shouldYield ifTrue: [^ self yield].
	expression _ stackFrame expression.

	(expression isKindOf: Symbol) ifTrue: [
		(#closeCustomBlock = expression) ifTrue: [
		^self closeCustomBlock]].

	stackFrame shouldUnlight ifTrue: [
		expression litUp: false.
		^ self popStackFrame].

	(expression isKindOf: ReporterScriptBlockMorph) ifTrue: [
		self returnValueToParentFrame: expression evaluate.
		^ self popStackFrame ].

	expression isCollection ifTrue: [^ self evaluateSequence].
	expression isArgMorph ifTrue: [^ self evaluateSelfEvaluating].
	expression isBlockMorph ifTrue: [^ self evaluateCommandFor: aStageMorph].

	self error: 'Unknown expression type:  ', expression printString.
! !

!ScratchProcess methodsFor: 'private-evaluation' stamp: 'jens 7/23/2010 01:15'!
evaluateNextArgument
	"Evaluates the next argument of the command in expression.  Requires that expression be a block that takes more args than are already evaluated in the current stack frame."

	| argumentExpression |

	argumentExpression _ stackFrame expression argumentAt: stackFrame arguments size + 1. 

	stackFrame expression selector = #doCustomBlock ifTrue: [
		((argumentExpression isKindOf: ReporterBlockMorph) and: [
			#(unevaluated unevaluatedBoolean) includes: argumentExpression declaredType]) ifTrue: [
				argumentExpression _ argumentExpression asLambda ]].

	self pushStackFrame: (ScratchStackFrame new expression: argumentExpression).! !

!ScratchProcess methodsFor: 'private-evaluation' stamp: 'jm 3/23/2005 09:53'!
evaluateSelfEvaluating
	"The easiest case. Evaluate a number, color, or any self-evaluating expression."

	| value |
	value _ stackFrame expression evaluate.
	self returnValueToParentFrame: value.
	self popStackFrame.
! !

!ScratchProcess methodsFor: 'private-evaluation' stamp: 'DaveF 7/9/2003 10:45'!
evaluateSequence
	"Requires that the current expression be a collection of blocks.  Evaluates the block in the current expression (which is a list of blocks) at the given pc."

	| blocks pc |

	blocks _ stackFrame expression.
	pc _ stackFrame pc.

	pc > blocks size
		ifTrue:	[self popStackFrame]
		ifFalse:	[stackFrame pc: pc + 1.
				self pushStackFrame: (ScratchStackFrame new expression: (blocks at: pc))].

! !

!ScratchProcess methodsFor: 'private-evaluation' stamp: 'jens 7/28/2010 20:14'!
evaluateSpecialForm
	"Evaluates the current special form expression.  Requires that no arguments have been evaluated, and that the current expression be a special form."

	|sel|

	stackFrame ifNil: [^ self].
	sel _ stackFrame expression selector.
	sel ifNil: [sel _ #nop].
	(#(getBlockVar: getVar:) includes: sel) ifTrue: [^self getVar].
	(#(setBlockVar:to: setVar:to:) includes: sel) ifTrue: [^self setVar].
	(#(changeBlockVar:by: changeVar:by:) includes: sel) ifTrue: [^self changeVar].

	self perform: sel.! !

!ScratchProcess methodsFor: 'private-evaluation' stamp: 'jens 5/18/2010 00:05'!
yield
	"Sets a flag indicating that we're ready to yield to another process."

	readyToYield _ true.

	isPaused ifFalse: [
		self popStackFrame ].! !


!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 7/26/2010 22:53'!
closeCustomBlock

	blockVarList ifNotNil: [  blockVarList size > 0 ifTrue: [
		blockVarList removeLast.
		atomicityList removeLast.
		scriptList removeLast ]].
	stackFrame ifNotNil: [self popStackFrame ].

	nextBlockToDebug ifNotNil: [self nextStep]
	
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 8/10/2010 23:34'!
doAnswer
	"Evaluates its argument, and returns the value to the frame from which the current method was called."

	| value args zap |

	args _ stackFrame arguments.

	"Evaluate the argument, if necessary."

	args size < stackFrame expression argumentCount
		ifTrue: [^self evaluateNextArgument].


	"Remember the return value."
	args size > 0
		ifTrue:	[value _ args first]
		ifFalse:	[value _ nil].
	answer _ value. 

	zap _ [
		[ stackFrame isNil
		or: [stackFrame expression = #closeCustomBlock ]] whileFalse: [ 
			self popStackFrame]. self closeCustomBlock].

	zap value.

	"get out of a custom C-shape command slot"
	[stackFrame notNil 
	and: [(stackFrame expression isKindOf: Array)
		and: [stackFrame expression first isKindOf: MultiArgCommandBlockMorph]]] whileTrue: [ 
		zap value. zap value]. 

	[stackFrame notNil 
	and: [stackFrame parentFrame notNil 
	and: [stackFrame parentFrame expression = #shouldYield]]] whileTrue: [
		zap value. zap value ].

	self returnValueToParentFrame: value.

	stackFrame ifNotNil: [ self popStackFrame ].
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 4/19/2011 17:37'!
doApplyEval
	"execute a user-defined block"

	| block script lambda varsFrame parms emptySlots nestedLambdas argsList stage stopFrame |

	self isAtomic. "allow escaping out of infinite recursion"
	block _ stackFrame expression.

	"evaluate arguments, if necessary"

	stackFrame arguments size < block argumentCount
		ifTrue: [^ self evaluateNextArgument].

	#doCustomBlock = block selector
		ifTrue: [
			lambda _ block.
			argsList _ stackFrame arguments]
		ifFalse: [
			lambda _ stackFrame arguments first.
			argsList _ #().
			stackFrame arguments size > 2  ifTrue: [
				(block selector endsWith: 'BlockWithArgList') ifTrue: [
					argsList _ stackFrame arguments third asArray].
				(block selector endsWith: 'BlockWithArgs') ifTrue: [
					argsList _ stackFrame arguments copyFrom: 3 to: stackFrame arguments size]]].

	script _ lambda sequence collect: [:eachBlock | eachBlock fullCopy].
	atomicityList ifNil: [atomicityList _ OrderedCollection new].
	atomicityList add: lambda isAtomic.
	blockVarList ifNil: [blockVarList _ OrderedCollection new].
	blockVarList add: (VariableFrame forNames: lambda allVarNames).
	varsFrame _ blockVarList last.


	"Step 0:  If the block being called provides explicit formal parameters
	(INPUT NAMES), then replace every formal parameter reference in the body
	with the corresponding actual argument value.  Do not do any automatic
	substitution of arguments into empty/default slots.  Otherwise:

	Step 1:  If the total number of empty + default input slots in the
	block being called is exactly equal to the number of actual arguments
	provided (no special treatment for explicit empty actual arguments -- it's
	the empty string as the argument), 	then fill every empty + default slot 
	with the corresponding actual argument.

	Step 2:  If step 1 fails, but the total number of empty (not default) input
	slots in the block being called is exactly equal to the number of actual
	arguments provided, then fill every empty (not default) slot with the
	corresponding actual argument.

	Step 3:  If steps 1 and 2 fail, but the number of actual arguments
	provided is 1, then fill every empty (not default) slot with that argument.

	Step 4:  If steps 1-3 fail, do not do any automatic substitution at all.

	Variadic inputs"


 	"bind formal parameters"

	parms _ lambda parameterNames.
	parms size > 0
		ifTrue: [
			1 to: parms size do: [:i|
				argsList size >= i ifTrue: [
					varsFrame setVar: (parms at: i) to: (argsList at: i)]]]

		ifFalse: [ argsList size > 0 ifTrue: [

			"distribute implicit parameters"

			"collect all empty slots and nested lambdas"

			emptySlots _ OrderedCollection new.
			"emptyDefaults _ OrderedCollection new."
			nestedLambdas _ OrderedCollection new.
			script do: [:eachBlock |
				eachBlock allMorphs do: [:m |
					(m isKindOf: ArgMorph) 
						ifTrue: [
							m isEmptySlot ifTrue: [
								emptySlots add: m.

"--- rule 1 commented out for now -------------------------------------------------------------------------------------------------

								default _ m defaultValue.
								default ifNotNil: [
									emptyDefaults add: m ]

------------------------------------------------------------------------------------------------------------------------------------------"

															]]
						ifFalse: [ (m isKindOf: LambdaBlockMorph)
							ifTrue: [nestedLambdas add: m]]]].

			"remove slots in nested lambdas from list"
			nestedLambdas do: [:l |
				emptySlots removeAll: l allMorphs.

"--- rule 1 commented out for now -------------------------------------------------------------------------------------------------

				emptyDefaults removeAll: l allMorphs 

------------------------------------------------------------------------------------------------------------------------------------------"


														].


"--- rule 1 commented out for now -------------------------------------------------------------------------------------------------

			emptyDefaults size = argsList size
				ifTrue: [
	
					""fill every empty+default slot with the corresponding actual argument,
					substitute empty string for default value""
	
					1 to: emptySlots size do: [:i |
						| obj slot |
						slot _ emptySlots at: i.
	
						""empty variadic inputs absorb all ensuing values""
						(slot isKindOf: MultiArgMorph) 
							ifTrue: [ obj _ ScratchListMorph on: argsList from: i to: argsList size]
							ifFalse: [
								obj _ argsList at: i.
								(obj isKindOf: ReporterBlockMorph)
									ifTrue: [ obj _ obj evaluate ]
									ifFalse: [ obj = '' asUTF8 
										ifTrue: [ obj _ slot defaultValue]]].
						slot implicitValue: obj ]]

				ifFalse: [


------------------------------------------------------------------------------------------------------------------------------------------"

						 (emptySlots size = argsList size or: [argsList size > emptySlots size and: [emptySlots size > 0 and: [emptySlots last isKindOf: MultiArgMorph]]])
					ifTrue: [
						"fill every empty (not default) slot with the corresponding actual argument"
	
						1 to: emptySlots size do: [:i |
							| obj slot |
							slot _ emptySlots at: i.
		
							"empty variadic inputs absorb all ensuing values"
							(slot isKindOf: MultiArgMorph) 
								ifTrue: [ obj _ ScratchListMorph on: argsList from: i to: argsList size]
								ifFalse: [
									obj _ argsList at: i.
									(obj isKindOf: ReporterBlockMorph)
										ifTrue: [ obj _ obj evaluate ]].
							slot implicitValue: obj ]]
	
					ifFalse: [argsList size = 1 ifTrue: [

						"put the single parameter into every empty slot"

						| prm |
						prm _ argsList first.
						(prm isKindOf: ReporterBlockMorph)
							ifTrue: [prm _ prm evaluate ].
						emptySlots do: [:slot|
							(slot isKindOf: MultiArgMorph)
								ifTrue: [slot implicitValue: (ScratchListMorph with: prm)]
								ifFalse: [slot implicitValue: prm]]]]]"]"].

	scriptList ifNil: [scriptList _ OrderedCollection new].
	scriptList add: script.

	self bindVarsIn: script to: varsFrame.

	"bind externalized local vars"

	#doCustomBlock = block selector ifTrue: [
		| templates aliases argMorphs |

		templates _ block definition templates.
		templates ifNotNil: [
			argMorphs _ block argMorphs.
			aliases _ Dictionary new.
			templates keysDo: [:i |
				aliases at: (argMorphs at: i) variable put: (templates at: i)].
			argsList do: [:eachArg | (eachArg isKindOf: LambdaBlockMorph) ifTrue: [
				aliases keysDo: [:vName |
					eachArg bind: vName to: varsFrame as: (aliases at: vName)]]]]].

	(block isReporter and: [lambda isFunction]) ifFalse: [self popStackFrame].

	(block selector beginsWith: 'doFork')
		ifTrue: [
			stage _ block receiver ownerThatIsA: ScratchStageMorph.
			"self popStackFrame."
			stage startProcessForSequence: scriptList]
		ifFalse: [
			stopFrame _ ScratchStackFrame new expression: #closeCustomBlock.
			self pushStackFrame: stopFrame.
			(#(doCustomBlock doAnswer) includes: block selector) ifTrue: [
				self bindStopBlocksIn: script to: stopFrame].
			self pushStackFrame: (ScratchStackFrame new expression: script)]! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 7/29/2010 00:24'!
doAsk
	"Ask the user a question and wait until the user types some text and hits the accept button."

	| block args question prompter |

	block _ stackFrame expression.
	args _ stackFrame arguments.

	"first time (args size = 0): open prompter"
	args size = 0
		ifTrue: [
			block receiver promptInProgress ifTrue: [
				self pushStackFrame: (ScratchStackFrame new shouldYield: true).
				^ self].
			question _ block args first "asString".
			prompter _ block receiver promptForInput: question]
		ifFalse: [
			prompter _ args first].

	"if all processes have stopped, pop this frame and we're done."
	prompter isDone ifTrue: [^ self popStackFrame].

	"Set up the wait loop. We need to:
		1.  yield to another process
		2.  evaluate the termination expression again
	Push these onto stack in reverse order."

	"2.  evaluate this block again"
	"(Remove the current frame and replace it with one that will force the termination expression to be re-evaluated.)"

	self popStackFrameNoUnhightlight.
	self pushStackFrame: (ScratchStackFrame new expression: block; addArgument: prompter).

	"1.  yield to another process"
	self pushStackFrame: (ScratchStackFrame new shouldYield: true).
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 11/17/2009 21:09'!
doBroadcastAndWait
	"Broadcast an event and wait until all processes started by that broadcast have terminated."

	| block args eventName stage procs arg vars|
	block _ stackFrame expression.
	args _ stackFrame arguments.

	"first time: send the event and collect a list of processes"
	args size = 0
		ifTrue: [
			eventName _ block args first asString.
			eventName = '0' ifTrue: [
				arg _ block argMorphs first.
				blockVarList ifNotNil: [blockVarList isEmpty ifFalse: [
					vars _ blockVarList last.
					eventName _ (vars getVar: arg variable) asString ]]].
			stage _ block receiver ownerThatIsA: ScratchStageMorph.
			procs _ stage broadcastEventNamed: eventName with: 0]
		ifFalse: [
			procs _ args first].

	"if all processes have stopped, pop this frame and we're done."
	(procs select: [:p | p isRunning]) size = 0 ifTrue: [^ self popStackFrame].

	"Set up the wait loop. We need to:
		1.  yield to another process
		2.  evaluate the termination expression again
	Push these onto stack in reverse order."

	"2.  evaluate this block again"
	"(Remove the current frame and replace it with one that will force the termination expression to be re-evaluated.)"
	self popStackFrameNoUnhightlight.
	self pushStackFrame: (ScratchStackFrame new expression: block; addArgument: procs).

	"1.  yield to another process"
	self pushStackFrame: (ScratchStackFrame new shouldYield: true).
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 2/22/2010 00:43'!
doCallBlockWithArgList
	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 1/5/2010 04:29'!
doCallBlockWithArgs
	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 1/5/2010 04:27'!
doCustomBlock
	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 9/22/2010 01:13'!
doDeclareVariables

	| block env  |

	block _ stackFrame expression.
	(blockVarList notNil and: [blockVarList size > 0])
		ifTrue: [ env _ blockVarList last ]
		ifFalse: [ env _ tempVars ].
	block variables do: [:var | env add: var ]. 
	(scriptList notNil and: [scriptList size > 0])
		ifTrue: [self bindVarsIn: scriptList last to: env]
		ifFalse: [self bindVarsIn: block blockSequence to: env].

	self popStackFrame! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jm 8/24/2003 12:12'!
doForever
	"Executes one iteration of a forever loop."

	"Right now, the forever special form is at the top of the stack.  We're going to add two more things on top of this:  (1) a frame for evaluating the blocks in the forever block, and (2) a frame that just results in yielding control.  That way we'll evaluate the contents of the block, give up control, and do it all again."

	| blocks |

	"Remember blocks to evaluate before we clobber the stack."
	blocks _ stackFrame expression firstBlockList.	

	"When we're all done with these blocks, we want to yield."
	self pushStackFrame: (ScratchStackFrame new shouldYield: true).

	"First, we want to evaluate the blocks inside the loop."
	self pushStackFrame: (ScratchStackFrame new expression: blocks).
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 12/9/2009 13:00'!
doForeverIf
	"Executes one iteration of a forever loop."

	| block args |
	block _ stackFrame expression.
	args _ stackFrame arguments.

	"evaluate the condition if we haven't yet"
	args size < 1 ifTrue: [^ self evaluateNextArgument].

	"remove the current frame and replace it with one that will force the test expression to be re-evaluated."
	self popStackFrameNoUnhightlight.
	self pushStackFrame: (ScratchStackFrame new expression: block).
	self pushStackFrame: (ScratchStackFrame new shouldYield: true). "yield after possibly running body"

	"if the condition was true, run the body"
	args first = true ifTrue: [
		self pushStackFrame: (ScratchStackFrame new expression: block firstBlockList)].

! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 2/23/2010 00:44'!
doFork

	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 3/1/2011 18:17'!
doForkBlockWithArgList
	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 2/23/2010 00:45'!
doForkBlockWithArgs
	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 12/9/2009 12:55'!
doIf
	"Evaluate the current expression (which must be an if)."

	| block arguments argExp |

	block _ stackFrame expression.
	arguments _ stackFrame arguments.

	"Evaluate the arg if we haven't already."
	arguments size = 0
		ifTrue:	[argExp _ block argumentAt: 1.
				^self pushStackFrame: (ScratchStackFrame new expression: argExp)].

	"We can pop this expression off the stack either way."
	self popStackFrame.

	"If the predicate was false, just return."
	arguments first = true
		ifFalse: [^self].

	"Since the predicate was true, evaluate the body of the if."
	self pushStackFrame: (ScratchStackFrame new expression: block firstBlockList).! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 12/9/2009 12:57'!
doIfElse
	"Evaluate the current expression (which must be an if-then-else)."

	| block arguments argExp |
	block _ stackFrame expression.
	arguments _ stackFrame arguments.

	"evaluate the arg if we haven't already."
	arguments size = 0 ifTrue: [
		argExp _ block argumentAt: 1.
		^ self pushStackFrame: (ScratchStackFrame new expression: argExp)].

	"we can pop this expression off the stack either way"
	self popStackFrame.

	arguments first = true
		ifTrue: [self pushStackFrame: (ScratchStackFrame new expression: block trueBlockList)]
		ifFalse: [self pushStackFrame: (ScratchStackFrame new expression: block falseBlockList)].
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 4/28/2010 00:07'!
doMakeNewList
	| block list |

	block _ stackFrame expression.

	"evaluate arguments, if necessary"
	stackFrame arguments size < block argumentCount
		ifTrue: [^ self evaluateNextArgument].

	list _ ScratchListMorph new.
	list listName: '' target: nil.
	stackFrame arguments do: [:arg |	
		list insertLine: (ScriptableScratchMorph new asListElement: arg) at: (list lineCount + 1) ].
	list updateCountMorph.
	list fixLayout.

	answer _ list.
	self returnValueToParentFrame: list.
	self popStackFrame
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 7/8/2010 20:39'!
doPlaySoundAndWait
	"Play a sound and block until it has finished playing."

	| block args soundName snd |
	block _ stackFrame expression.
	args _ stackFrame arguments.

	"first time: send the event and collect a list of processes"
	args size = 0
		ifTrue: [
			soundName _ block args first.
			snd _ block receiver soundNamed: soundName ifAbsent: [^ self popStackFrame].
			snd playFromStart]
		ifFalse: [
			snd _ args first].

	"if sound has stopped, pop this frame and we're done"
	snd isPlaying ifFalse: [ ^ self popStackFrame].

	"Set up the wait loop. We need to:
		1.  yield to another process
		2.  evaluate our termination condition again
	Push these onto stack in reverse order."

	"2.  evaluate this block again"
	"(Remove the current frame and replace it with one that will force the termination condition to be re-evaluated.)"
	self popStackFrameNoUnhightlight.
	self pushStackFrame: (ScratchStackFrame new
		expression: block;
		addArgument: snd).

	"1.  yield to another process"
	self pushStackFrame: (ScratchStackFrame new shouldYield: true).
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 9/26/2008 12:34'!
doRepeat
	"Handles one iteration of a repeat block."

	| arguments argExp block counter frame |

	block _ stackFrame expression.
	arguments _ stackFrame arguments.

	"If we haven't done so yet, evaluate the argument to repeat."
	arguments size < 1
		ifTrue:	[argExp _ block argumentAt: 1.
				^self pushStackFrame: (ScratchStackFrame new expression: argExp)].

	"If the number of times to repeat is 0, then we're done."
	counter _ arguments first asNumberNoError.
	counter <= 0 ifTrue:	[^ self popStackFrame].

	"Pop this instruction from the stack."
	self popStackFrameNoUnhightlight.

	"At top of stack should now be:
		1.  evaluate body of repeat block.
		2.  yield.
		3.  evaluate repeat block with decremented counter value.
	Need to add these to the stack in reverse order."

	"3.  evaluate repeat block with decremented counter value."
	frame _ ScratchStackFrame new
				expression:		block;
				addArgument:	counter - 1.
	self pushStackFrame: frame.

	"2.  yield."
	self isAtomic ifFalse: [ self pushStackFrame: (ScratchStackFrame new shouldYield: true) ].

	"1.  evaluate body of repeat block."
	self pushStackFrame: (ScratchStackFrame new expression: block firstBlockList).
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 6/18/2010 01:21'!
doReport
	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jm 3/22/2005 11:17'!
doReturn
	"Evaluates its argument, and returns the value to the frame from which the current method was called."

	| value args |

	args _ stackFrame arguments.

	"Evaluate the argument, if necessary."
	args size < stackFrame expression argumentCount
		ifTrue: [^self evaluateNextArgument].

	"Remember the return value."
	args size > 0
		ifTrue:	[value _ args first]
		ifFalse:	[value _ nil].

	"Pop until we're out of frames to pop, or we hit a return marker."

	[stackFrame isNil] whileFalse:
		[self popStackFrame].

	stackFrame ifNotNil:	[
		self returnValueToParentFrame: value.
		self popStackFrame].
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 1/5/2010 04:28'!
doRun
	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 2/22/2010 00:43'!
doRunBlockWithArgList
	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 1/5/2010 04:28'!
doRunBlockWithArgs
	self doApplyEval! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 8/3/2010 12:46'!
doStopBlock
	"stops the current custom block"

	| stopFrame |
	stopFrame _ stackFrame expression stopFrame.

	self popStackFrame.

	[stackFrame parentFrame isNil or: [stackFrame == stopFrame]] whileFalse: [
		stackFrame _ stackFrame parentFrame].

	stackFrame ifNotNil: [
		self popStackFrame].

! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'DaveF 7/25/2003 11:15'!
doUntil
	"Evaluates the current until-loop expression."

	^self doUntil: true.! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 12/9/2009 13:03'!
doUntil: terminatingBoolean
	"Evaluates the current while or until loop expression until the loop condition evaluates to the boolean terminating value."

	| args block |

	block _ stackFrame expression.
	args _ stackFrame arguments.

	"Evaluate the condition if we haven't yet."
	args size < 1
		ifTrue: [^self evaluateNextArgument].

	"If the condition matches the terminatingBoolean, pop this frame and we're done."
	args first = terminatingBoolean = true ifTrue: [^self popStackFrame].

	"Set up for an iteration of the while loop.  We need to:
		1.  evaluate the body
		2.  yield to another process
		3.  evaluate the while/until block again
	Push these onto stack in reverse order."

	"3.  evaluate the while/until block again"
	"(We remove the current frame and replace it with one that will force the argument to be re-evaluated.)"
	self popStackFrameNoUnhightlight.
	self pushStackFrame: (ScratchStackFrame new expression: block).

	"2.  yield to another process"
	self isAtomic ifFalse: [self pushStackFrame: (ScratchStackFrame new shouldYield: true)].
	
	"1.  evaluate the body"
	self pushStackFrame: (ScratchStackFrame new expression: block firstBlockList).
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 12/9/2009 13:05'!
doWaitUntil
	"Evaluates the the wait condition until it evaluates to true."

	| block args |
	block _ stackFrame expression.
	args _ stackFrame arguments.

	"evaluate the termination expression if we haven't yet."
	args size < 1 ifTrue: [^ self evaluateNextArgument].

	"if the termination condition is met, pop this frame and we're done."
	args first = true ifTrue: [^ self popStackFrame].

	"Set up for waitUntil loop. We need to:
		1.  yield to another process
		2.  evaluate the termination expression again
	Push these onto stack in reverse order."

	"2.  evaluate this block again"
	"(Remove the current frame and replace it with one that will force the termination expression to be re-evaluated.)"
	self popStackFrameNoUnhightlight.
	self pushStackFrame: (ScratchStackFrame new expression: block).

	"1.  yield to another process"
	self pushStackFrame: (ScratchStackFrame new shouldYield: true).
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'DaveF 7/25/2003 11:14'!
doWhile
	"Evaluates the current while-loop expression."

	^self doUntil: false.! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 8/10/2010 15:36'!
xdoAnswer
	"Evaluates its argument, and returns the value to the frame from which the current method was called."

	| value args zap |

	args _ stackFrame arguments.

	"Evaluate the argument, if necessary."

	args size < stackFrame expression argumentCount
		ifTrue: [^self evaluateNextArgument].


	"Remember the return value."
	args size > 0
		ifTrue:	[value _ args first]
		ifFalse:	[value _ nil].
	answer _ value. 

	zap _ [
		[ stackFrame isNil
		or: [stackFrame expression = #closeCustomBlock ]] whileFalse: [ 
			self popStackFrame]. self closeCustomBlock].

	zap value.

	"get out of a custom C-shape command slot"
	[stackFrame notNil 
	and: [(stackFrame expression isKindOf: Array)
		and: [stackFrame expression first isKindOf: MultiArgCommandBlockMorph]]] whileTrue: [ 
		zap value. zap value]. 

	[stackFrame notNil 
	and: [stackFrame parentFrame notNil 
	and: [stackFrame parentFrame expression = #shouldYield]]] whileTrue: [
		zap value. zap value ].

	self returnValueToParentFrame: value.

	stackFrame ifNotNil: [ self popStackFrame ].
! !

!ScratchProcess methodsFor: 'private-special forms' stamp: 'jens 8/10/2010 15:46'!
xnewdoAnswer
	"Evaluates its argument, and returns the value to the frame from which the current method was called."

	| value args stopFrame zap |

	stopFrame _ stackFrame expression stopFrame.
	args _ stackFrame arguments.

	"Evaluate the argument, if necessary."

	args size < stackFrame expression argumentCount
		ifTrue: [^self evaluateNextArgument].

	"Remember the return value."
	args size > 0
		ifTrue:	[value _ args first]
		ifFalse:	[value _ nil].
	answer _ value. 

	zap _ [
		[ stackFrame isNil
		or: [stackFrame expression = #closeCustomBlock ]] whileFalse: [ 
			self popStackFrame]. self closeCustomBlock].

	[stackFrame isNil
	or: [stackFrame parentFrame isNil
	or: [stackFrame parentFrame parentFrame isNil
	or: [stackFrame == stopFrame]]]] whileFalse: [
			zap value].

	stackFrame ifNotNil: [
		self popStackFrame].

	self returnValueToParentFrame: value.

	stackFrame ifNotNil: [
		self popStackFrame].

! !


!ScratchProcess methodsFor: 'private-stack' stamp: 'jm 1/22/2007 13:12'!
popStackFrameNoUnhightlight
	"Pops the current stack frame off the top of the stack, so that the next one becomes the current frame. Do not add a stack frame to unhighlight this block. Use this method when a block will re-evaluated, such 'wait until'."

	stackFrame _ stackFrame parentFrame.

! !

!ScratchProcess methodsFor: 'private-stack' stamp: 'jm 3/23/2005 09:06'!
pushStackFrame: aScratchStackFrame
	"Pushes the given new stack frame onto the top of the stack, so that it becomes the current stack frame."

	aScratchStackFrame parentFrame: stackFrame.
	stackFrame _ aScratchStackFrame.
! !

!ScratchProcess methodsFor: 'private-stack' stamp: 'jens 7/22/2010 23:24'!
returnValueToParentFrame: anObject
	"Append the given object to the argument list of my parent frame."

	| f obj |
	obj _ anObject.
	(anObject isKindOf: LambdaBlockMorph) ifTrue: [
		obj _ anObject fullCopy.
		obj fixInitialFrame].
	stackFrame ifNil: [^ self].
	f _ stackFrame parentFrame.
	[f notNil and: [f shouldUnlight ] ] whileTrue: [
		f _ f parentFrame]. 

	f ifNotNil: [f addArgument: obj].

	((blockToDebug isKindOf: BlockMorph) and: [self currentBlock isKindOf: ReporterBlockMorph]) ifTrue: [
			((ScratchFrameMorph someInstance viewMode = #presentation) | doneWithDebugging) ifFalse: [
				self pause. 
				DebuggerFrameMorph new on: self returnValue: obj]].

! !

!ScratchProcess methodsFor: 'private-stack' stamp: 'DaveF 7/9/2003 18:26'!
stackAsArray
	"Returns an array describing the contents of the stack.  Useful for debugging."

	| frame collection |

	frame _ stackFrame.
	collection _ OrderedCollection new.
	
	[frame isNil] whileFalse:
		[collection add: frame expression.
		frame _ frame parentFrame].

	^collection asArray! !


!ScratchProcess methodsFor: 'BYOB-pause/resume' stamp: 'jens 5/21/2010 01:28'!
currentBlock
	stackFrame ifNil: [^nil].
	(stackFrame expression isKindOf: BlockMorph)
		ifFalse: [^nil].
	^ stackFrame expression! !

!ScratchProcess methodsFor: 'BYOB-pause/resume' stamp: 'jens 5/20/2010 16:13'!
doPauseThread

	self nextStep.
	self popStackFrame.

! !

!ScratchProcess methodsFor: 'BYOB-pause/resume' stamp: 'jens 7/27/2010 20:58'!
doPauseThreadReporter


	| value args |

	self nextStep.

	args _ stackFrame arguments.
	"Evaluate the argument, if necessary."
	args size < stackFrame expression argumentCount
		ifTrue: [ ^self evaluateNextArgument].

	"Remember the return value."
	args size > 0
		ifTrue:	[value _ args first]
		ifFalse:	[value _ nil].
	answer _ value. 

	self returnValueToParentFrame: value.

	stackFrame ifNotNil: [
		self popStackFrame]
! !

!ScratchProcess methodsFor: 'BYOB-pause/resume' stamp: 'jens 5/18/2010 00:02'!
isPaused
	^isPaused ifNil: [^ isPaused _ false]! !

!ScratchProcess methodsFor: 'BYOB-pause/resume' stamp: 'jens 6/10/2010 01:09'!
jumpOver
	| cb |
	cb _ self currentBlock.
	blockToDebug _ nil.
	nextBlockToDebug _ cb nextBlock.
	isPaused _ false! !

!ScratchProcess methodsFor: 'BYOB-pause/resume' stamp: 'jens 6/18/2010 01:53'!
nextStep

	blockToDebug _ self currentBlock. 
	nextBlockToDebug _ nil.
	isPaused _ false! !

!ScratchProcess methodsFor: 'BYOB-pause/resume' stamp: 'jens 7/8/2010 19:53'!
pause
	timeOffset _ nil.
	isPaused _ true! !

!ScratchProcess methodsFor: 'BYOB-pause/resume' stamp: 'jens 6/18/2010 01:56'!
resume

	blockToDebug _ nil.
	nextBlockToDebug _ nil.
	isPaused _ false.
	doneWithDebugging _ true
! !


!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 8/10/2010 14:57'!
bindStopBlocksIn: blockSequence to: aStackFrame

	blockSequence do: [:block |
		block allMorphsDo: [:m |
			((m isKindOf: CommandBlockMorph) and: [(#(doStopBlock doAnswer) includes: m selector )]) ifTrue: [
				m stopFrame: aStackFrame]]]! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 9/21/2010 23:33'!
bindVarsIn: blockSequence to: aVariableFrame

	blockSequence do: [:block |
		block allMorphsDo: [:m | 
			((m isKindOf: VariableBlockMorph) | (m isKindOf: SetterBlockMorph)
			and: [m isSpecialForm & (aVariableFrame scope includes: m variable)]) ifTrue: [
				m receiver: aVariableFrame ]]].! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 3/26/2010 00:56'!
changeVar

	| block arguments argExp environment |

	block _ stackFrame expression.
	arguments _ stackFrame arguments.

	"Evaluate the arg if we haven't already."
	arguments size = 0
		ifTrue:	[argExp _ block argumentAt: 2.
				^self pushStackFrame: (ScratchStackFrame new expression: argExp)].

	(block receiver isKindOf: VariableFrame) 
		ifTrue: [ environment _ block receiver ]
		ifFalse: [ (blockVarList notNil and: [blockVarList isEmpty not])
			ifTrue: [ environment _ blockVarList last ]].
	environment ifNotNil: [
		environment changeVar: (block argumentAt: 1) evaluate by: arguments first ].
	self popStackFrame.! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 5/20/2010 23:03'!
currentEnvironment
	"private"

	blockVarList ifNotNil: [
		blockVarList size > 0 ifTrue: [
			^ blockVarList last]].
	^ nil! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 1/18/2011 23:22'!
getVar

	| vars block value|
	block _ stackFrame expression.
	(block receiver isKindOf: VariableFrame)
		ifTrue: [value _ block evaluate]
		ifFalse: [
			value _ 0.
			blockVarList ifNotNil: [blockVarList isEmpty ifFalse: [
				vars _ blockVarList last.
				value _ vars getVar: block variable ]]]. 
	self returnValueToParentFrame: value. 
	self popStackFrame.! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 1/18/2011 23:20'!
getVarNamed: vName 
	"strictly dynamic without bindings!!"

	| env block |
	block _ stackFrame expression.
	(blockVarList notNil and: [blockVarList size > 0]) ifTrue: [
		env _ blockVarList last.
		(env scope includes: vName)
			ifTrue: [ ^ env getVar: vName ]].
	(tempVars scope includes: vName)
		ifTrue: [^ tempVars getVar: vName ].
	^ block receiver getVar: vName
! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 7/21/2010 00:27'!
isAtomic

	| stopButtonRect presenter |

	presenter _ World findA: ScratchPresenterMorph.

	presenter notNil
		ifTrue: [ stopButtonRect _ presenter buttons fourth bounds ]
		ifFalse: [ stopButtonRect _ (World topRight + (-42@48)) extent: (20@20)].

	(Sensor anyButtonPressed and: [stopButtonRect containsPoint: Sensor cursorPoint]) ifTrue: [
		self stop.
		ScratchFrameMorph allInstances do: [:sf| sf stopAll]].

	((Character value: 27) = Sensor keyboardPeek) ifTrue: [
			Sensor flushKeyboard.
			Preferences noviceMode ifFalse: [
				Display flash: stopButtonRect].
			^ false ].

	atomicityList ifNil: [^false].
	(atomicityList size = 0) ifTrue: [^false].
	Sensor flushKeyboard.
	^atomicityList last


! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 3/15/2010 22:40'!
nop

	self returnValueToParentFrame: 0.
	self popStackFrame.! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 9/22/2010 01:28'!
scriptList: anOrderedCollection

	scriptList _ anOrderedCollection! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 7/28/2010 23:56'!
setVar

	| block arguments argExp environment |

	block _ stackFrame expression.
	arguments _ stackFrame arguments.

	"Evaluate the arg if we haven't already."
	arguments size = 0
		ifTrue:	[argExp _ block argumentAt: 2.
				^self pushStackFrame: (ScratchStackFrame new expression: argExp)].

	(block receiver isKindOf: VariableFrame) 
		ifTrue: [ environment _ block receiver ]
		ifFalse: [ (blockVarList notNil and: [blockVarList isEmpty not])
			ifTrue: [ environment _ blockVarList last ]].

	environment ifNotNil: [
		environment setVar: (block argumentAt: 1) evaluate to: arguments first ].
	self popStackFrame.
! !

!ScratchProcess methodsFor: 'BYOB-variables' stamp: 'jens 1/26/2010 22:23'!
temporaryVariables
	^tempVars! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ScratchProcess class
	instanceVariableNames: ''!

!ScratchProcess class methodsFor: 'class initialization' stamp: 'jm 3/22/2005 11:00'!
initialize

	BlockHighlightMSecs _ 1.
! !


!ScratchProcess class methodsFor: 'class variables' stamp: 'jm 8/10/2004 12:45'!
blockHighlightMSecs

	^ BlockHighlightMSecs
! !

!ScratchProcess class methodsFor: 'class variables' stamp: 'jm 8/8/2003 14:02'!
blockHighlightMSecs: aNumber

	BlockHighlightMSecs _ aNumber.
! !


ScratchProcess initialize!
